/*
 *	Handle firewalling
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek               <buytenh@gnu.org>
 *	Bart de Schuymer		<bart.de.schuymer@pandora.be>
 *
 *	$Id: br_netfilter.c,v 1.17 2001/12/02 19:35:25 buytenh Exp $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/ip.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/if_ether.h>
#include <linux/netfilter_bridge.h>
#include <linux/netfilter_ipv4.h>
#include <net/ip.h>
#include <net/tcp.h>
#include <asm/uaccess.h>
#include <asm/checksum.h>
#include "br_private.h"


//#define dprintk printk
#define dprintk(args...)


#ifndef WE_REALLY_INSIST_ON_NOT_HAVING_NAT_SUPPORT
/* As the original source/destination addresses are variables private to this
 * file, we store them in unused space at the end of the control buffer.
 * On 64-bit platforms the TCP control buffer size still leaves us 8 bytes
 * of space at the end, so that fits.  Usage of the original source address
 * and the original destination address never overlaps (daddr is needed
 * around PRE_ROUTING, and saddr around POST_ROUTING), so that's okay as
 * well.
 */
#define skb_origaddr(skb)		(*((u32 *)(skb->cb + sizeof(skb->cb) - 4)))

#define store_orig_dstaddr(skb)		(skb_origaddr(skb) = skb->nh.iph->daddr)
#define store_orig_srcaddr(skb)		(skb_origaddr(skb) = skb->nh.iph->saddr)
#define dnat_took_place(skb)		(skb_origaddr(skb) != skb->nh.iph->daddr)
#define snat_took_place(skb)		(skb_origaddr(skb) != skb->nh.iph->saddr)
#else
#define store_orig_dstaddr(skb)
#define store_orig_srcaddr(skb)
#define dnat_took_place(skb)		(0)
#define snat_took_place(skb)		(0)
#endif


#define has_bridge_parent(device)	((device)->br_port != NULL)
#define bridge_parent(device)		(&((device)->br_port->br->dev))


static inline void __maybe_fixup_src_address(struct sk_buff *skb)
{
	if (snat_took_place(skb) &&
	    inet_addr_type(skb->nh.iph->saddr) == RTN_LOCAL) {
		memcpy(skb->mac.ethernet->h_source,
			bridge_parent(skb->dev)->dev_addr,
			ETH_ALEN);
	}
}



/* PF_BRIDGE/PRE_ROUTING *********************************************/
static struct net_device __fake_net_device = {
	hard_header_len:	ETH_HLEN
};

static struct dst_entry __fake_dst_entry = {
	__refcnt:		ATOMIC_INIT(1),
	dev:			&__fake_net_device,
	pmtu:			1500
};


static int br_nf_pre_routing_finish(struct sk_buff *skb)
{
	struct net_device *dev = skb->dev;
	struct iphdr *iph = skb->nh.iph;

	dprintk(KERN_NOTICE "br_nf_pre_routing_finish.. ");

	if (dnat_took_place(skb)) {
		if (ip_route_input(skb, iph->daddr, iph->saddr, iph->tos, dev)) {
			struct rtable *rt;

			if (ip_route_output(&rt, iph->daddr, 0, iph->tos, dev->ifindex)) {
				if (!ip_route_output(&rt, iph->daddr, 0, iph->tos, 0))
					printk(KERN_WARNING "br_nf_pre_routing_finish: shouldn't happen!\n");

				dprintk("must drop\n");
				kfree_skb(skb);
				return 0;
			}

			dprintk("divert\n");
			skb->dst = (struct dst_entry *)rt;
			memset(skb->mac.ethernet->h_dest, 0x22, ETH_ALEN);
		} else {
			dprintk("local ");
			memcpy(skb->mac.ethernet->h_dest, dev->dev_addr, ETH_ALEN);
		}
	} else {
		skb->dst = &__fake_dst_entry;
		dst_hold(skb->dst);
	}

	dprintk("\n");

	skb->dev = skb->physindev;
	br_handle_frame_finish(skb);

	return 0;
}

/* (PRE) okfn == br_handle_frame_finish */
static unsigned int br_nf_pre_routing(unsigned int hook, struct sk_buff **pskb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	struct iphdr *iph;
	__u32 len;
	struct sk_buff *skb;

	if ((*pskb)->mac.ethernet->h_proto != __constant_htons(ETH_P_IP))
		return NF_ACCEPT;

	dprintk(KERN_NOTICE "br_nf_pre_routing.. ");
	if ((skb = skb_share_check(*pskb, GFP_ATOMIC)) == NULL)
		goto out;

	if (!pskb_may_pull(skb, sizeof(struct iphdr)))
		goto inhdr_error;

	iph = skb->nh.iph;
	if (iph->ihl < 5 || iph->version != 4)
		goto inhdr_error;

	if (!pskb_may_pull(skb, 4*iph->ihl))
		goto inhdr_error;

	if (ip_fast_csum((__u8 *)iph, iph->ihl) != 0)
		goto inhdr_error;

	len = ntohs(iph->tot_len);
	if (skb->len < len || len < 4*iph->ihl)
		goto inhdr_error;

	if (skb->len > len) {
		__pskb_trim(skb, len);
		if (skb->ip_summed == CHECKSUM_HW)
			skb->ip_summed = CHECKSUM_NONE;
	}

	skb->physindev = skb->dev;
	skb->dev = bridge_parent(skb->dev);
	store_orig_dstaddr(skb);
	dprintk("calling br_nf_pre_routing_finish\n");
	NF_HOOK(PF_INET, NF_IP_PRE_ROUTING, skb, skb->dev, NULL,
		br_nf_pre_routing_finish);

	return NF_STOLEN;

inhdr_error:
//	IP_INC_STATS_BH(IpInHdrErrors);
out:
	dprintk("error out\n");
	return NF_DROP;
}


/* PF_BRIDGE/LOCAL_IN ************************************************/
static unsigned int br_nf_local_in(unsigned int hook, struct sk_buff **pskb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	struct sk_buff *skb = *pskb;

	if (skb->mac.ethernet->h_proto != __constant_htons(ETH_P_IP))
		return NF_ACCEPT;

	dprintk(KERN_NOTICE "br_nf_local_in..\n");
	if (skb->dst == &__fake_dst_entry) {
		dst_release(&__fake_dst_entry);
		skb->dst = NULL;
	}

	return NF_ACCEPT;
}


/* PF_BRIDGE/FORWARD *************************************************/
static int br_nf_forward_finish(struct sk_buff *skb)
{
	dprintk(KERN_NOTICE "br_nf_forward_finish..\n");
	__maybe_fixup_src_address(skb);
	br_forward_finish(skb);

	return 0;
}

static int br_nf_forward_postrouting(struct sk_buff *skb)
{
	dprintk(KERN_NOTICE "br_nf_forward_postrouting..\n");
	store_orig_srcaddr(skb);

	NF_HOOK(PF_INET, NF_IP_POST_ROUTING, skb, NULL,
		bridge_parent(skb->dev), br_nf_forward_finish);

	return 0;
}

static inline int output_maybe_reroute(struct sk_buff *skb)
{
	dprintk(KERN_NOTICE "output_maybe_reroute\n");

	/* @@@ this changes source MAC */
	return skb->dst->output(skb);
}

/* (PRE) okfn == br_forward_finish */
static unsigned int br_nf_forward(unsigned int hook, struct sk_buff **pskb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	struct sk_buff *skb = *pskb;
	int (*outfn)(struct sk_buff *);

	if (skb->mac.ethernet->h_proto != __constant_htons(ETH_P_IP))
		return NF_ACCEPT;

	dprintk(KERN_NOTICE "br_nf_forward.. ");

	skb->physoutdev = skb->dev;

	outfn = br_nf_forward_postrouting;
	if (dnat_took_place(skb))
		outfn = output_maybe_reroute;

	dprintk("call with outfn==%s\n", outfn==output_maybe_reroute?"omr":"bnfp");

	NF_HOOK(PF_INET, NF_IP_FORWARD, skb, bridge_parent(skb->physindev),
			bridge_parent(skb->dev), outfn);

	return NF_STOLEN;
}


/* PF_BRIDGE/LOCAL_OUT ***********************************************/
static int br_nf_local_out_finish(struct sk_buff *skb)
{
	dprintk(KERN_NOTICE "br_nf_local_out_finish..\n");
	__maybe_fixup_src_address(skb);
	br_forward_finish(skb);

	return 0;
}

static int br_nf_local_out_postrouting(struct sk_buff *skb)
{
	dprintk(KERN_NOTICE "br_nf_local_out_postrouting..\n");
	store_orig_srcaddr(skb);

	NF_HOOK(PF_INET, NF_IP_POST_ROUTING, skb, NULL,
		bridge_parent(skb->dev), br_nf_local_out_finish);

	return 0;
}

/* (PRE) okfn == br_forward_finish */
static unsigned int br_nf_local_out(unsigned int hook, struct sk_buff **pskb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	int hookno;
	struct net_device *realindev;
	struct sk_buff *skb = *pskb;

	if (skb->mac.ethernet->h_proto != __constant_htons(ETH_P_IP))
		return NF_ACCEPT;

	dprintk(KERN_NOTICE "br_nf_local_out.. ");

	skb->physoutdev = skb->dev;

	hookno = NF_IP_LOCAL_OUT;
	if ((realindev = skb->physindev) != NULL) {
		hookno = NF_IP_FORWARD;
		if (has_bridge_parent(realindev))
			realindev = bridge_parent(realindev);
	}

	dprintk("%s\n", hookno==NF_IP_LOCAL_OUT?"LO":"FW");

	NF_HOOK_THRESH(PF_INET, hookno, skb, realindev,
			bridge_parent(skb->dev),
			br_nf_local_out_postrouting,
			NF_IP_PRI_BRIDGE_SABOTAGE + 1);

	return NF_STOLEN;
}


/* IPv4/SABOTAGE *****************************************************/
static unsigned int ipv4_sabotage_in(unsigned int hook, struct sk_buff **pskb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	if (in->hard_start_xmit == br_dev_xmit &&
	    okfn != br_nf_pre_routing_finish) {
		struct sk_buff *skb = *pskb;

		dprintk(KERN_NOTICE "ipv4_sabotage_in %i %p\n", hook, skb);
		okfn(skb);
		return NF_STOLEN;
	}

	return NF_ACCEPT;
}

static unsigned int ipv4_sabotage_out(unsigned int hook, struct sk_buff **pskb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	if (out->hard_start_xmit == br_dev_xmit &&
	    okfn != br_nf_forward_postrouting &&
	    okfn != br_nf_forward_finish &&
	    okfn != br_nf_local_out_postrouting &&
	    okfn != br_nf_local_out_finish) {
		struct sk_buff *skb = *pskb;

		dprintk(KERN_NOTICE "ipv4_sabotage_out %i %p\n", hook, skb);
		if (hook == NF_IP_FORWARD && skb->physindev == NULL)
			skb->physindev = (struct net_device *)in;
		okfn(skb);
		return NF_STOLEN;
	}

	return NF_ACCEPT;
}


static struct nf_hook_ops br_nf_ops[] = {
	{ { NULL, NULL }, br_nf_pre_routing, PF_BRIDGE, NF_BR_PRE_ROUTING, 0 },
	{ { NULL, NULL }, br_nf_local_in, PF_BRIDGE, NF_BR_LOCAL_IN, 0 },
	{ { NULL, NULL }, br_nf_forward, PF_BRIDGE, NF_BR_FORWARD, 0 },
	{ { NULL, NULL }, br_nf_local_out, PF_BRIDGE, NF_BR_LOCAL_OUT, INT_MIN },
	{ { NULL, NULL }, ipv4_sabotage_in, PF_INET, NF_IP_PRE_ROUTING, INT_MIN },
	{ { NULL, NULL }, ipv4_sabotage_out, PF_INET, NF_IP_FORWARD, NF_IP_PRI_BRIDGE_SABOTAGE },
	{ { NULL, NULL }, ipv4_sabotage_out, PF_INET, NF_IP_LOCAL_OUT, NF_IP_PRI_BRIDGE_SABOTAGE },
	{ { NULL, NULL }, ipv4_sabotage_out, PF_INET, NF_IP_POST_ROUTING, INT_MIN },
};

#define NUMHOOKS (sizeof(br_nf_ops)/sizeof(br_nf_ops[0]))


int br_netfilter_init(void)
{
	int i;

#ifndef WE_REALLY_INSIST_ON_NOT_HAVING_NAT_SUPPORT
	struct sk_buff sk;

	if (sizeof(struct tcp_skb_cb) + 4 >= sizeof(sk.cb)) {
		extern int __too_little_space_in_control_buffer(void);
		__too_little_space_in_control_buffer();
	}
#endif

	for (i=0;i<NUMHOOKS;i++) {
		int ret;

		if ((ret = nf_register_hook(&br_nf_ops[i])) >= 0)
			continue;

		while (i--)
			nf_unregister_hook(&br_nf_ops[i]);

		return ret;
	}

	printk(KERN_NOTICE "Bridge firewalling registered\n");

	return 0;
}

void br_netfilter_fini(void)
{
	int i;

	for (i=NUMHOOKS-1;i>=0;i--)
		nf_unregister_hook(&br_nf_ops[i]);
}
